//C stuff
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#ifdef WIN32

/* windows part */
#undef SIGNAL
#include <windows.h>
#define SLEEP(x) Sleep(x)
#define GETPID() GetCurrentProcessId()
#else
/* unix part */
#define SIGNAL
#include <unistd.h>
#define SLEEP(x) sleep(x)
#define GETPID() getpid()
#endif
//STL
#include <vector>
#include <list>
#include <utility>
#include <algorithm>
#include <iterator>
//STD
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
//STR
#include <errno.h>
using namespace std;
#include "macro.h"
#include "automaton.h"
#include "seed.h"

#ifdef SIGNAL
#include <signal.h>
#endif

#ifndef PACKAGE
#define PACKAGE "iedera"
#endif

#ifndef VERSION
#define VERSION "1.02"
#endif

#define MAXSEEDS 32


#ifdef WIN32
/* windows part */
typedef unsigned __int64 QWORD;
QWORD GETMSTIME() {
  FILETIME filetime;
  GetSystemTimeAsFileTime(&filetime);
  return filetime.dwLowDateTime + filetime.dwHighDateTime * 1000000;
}
#else
/* unix part */
unsigned long long GETMSTIME() {
  struct timeval tv; 
  gettimeofday(&tv, 0);
  return tv.tv_usec + tv.tv_sec * 1000000;
}
#endif



// choose between a subset seed and a vectorized subset seed
#define SEEDAUTOMATON(autom,seed,nomerge)                   \
  ( gv_vectorized_flag ?                                    \
     (autom)->Automaton_SeedScore(                          \
               *(seed),                                     \
               gv_subsetseed_matching_matrix,               \
               gv_vectorizedsubsetseed_scoring_matrix,      \
               gv_vectorizedsubsetseed_scoring_threshold,   \
               nomerge                                      \
     )                                                      \
     :                                                      \
     (autom)->Automaton_SeedPrefixesMatching(               \
               *(seed),                                     \
                gv_subsetseed_matching_matrix,              \
		nomerge                                     \
     )                                                      \
   )


// not initalized
vector< vector<int> >  gv_subsetseed_matching_matrix;
vector< vector<int> >  gv_vectorizedsubsetseed_scoring_matrix;
int                    gv_vectorizedsubsetseed_scoring_threshold = 0;

vector< double >       gv_bsens;
int                    gv_bsens_k = 0;
vector< double >       gv_bsel;
int                    gv_bsel_k = 0;


// parameters
int      gv_alignment_length = 16;

bool     gv_minimize_flag      = true;
int      gv_nbruns             = 0;
bool     gv_hillclimbing_flag  = false;
double   gv_hillclimbing_alpha = 5e-2;
int      gv_motif              = 0;

seed *   gv_seeds[MAXSEEDS];
int      gv_nbseeds   = 1;
double   gv_jive      = 0.0;

int      gv_cycle_phase[MAXSEEDS];
int      gv_cycle                 = 1;
bool     gv_cycle_flag            = false;
bool     gv_cycle_fixedphase_flag = false;

vector< int >  gv_homogeneous_scores;
int            gv_homogeneous_score = 1;
bool           gv_homogeneous_flag = false;


bool     gv_sleeping_flag   = false;
char   * gv_output_filename = NULL;
char   * gv_input_filenames[256] = {NULL};
int      gv_nb_input_filenames =   0;


int build_default_subsetseed_matching_matrix();
int build_default_vectorizedsubsetseed_scoring_matrix();
int build_default_probabilities();
void DISPLAYMATRIX( std::ostream & os, vector< vector<int> > & matrix );



void USAGE() {
  cerr << "* USAGE:" << endl;
  cerr << "  " << PACKAGE << " [options]" << endl;
  cerr << "      -h       display this Help screen" << endl ;
  cerr << "      -v       display program Version" << endl ;
  cerr << "                                                                                      " << endl;
  cerr << "   1) Seed Model:" << endl;
  cerr << "     a) Data "<< endl;
  cerr << "      -A <int>                : align alphabet size \'A\' (default = " << gv_align_alphabet_size << ")" << endl;
  cerr << "      -B <int>                :  seed alphabet size \'B\' (default = " <<  gv_seed_alphabet_size << ")" << endl;
  cerr << "      -M {{<bool>}} table     : subset seed Matching matrix" << endl;
  cerr << "         (default = "; DISPLAYMATRIX(cerr,gv_subsetseed_matching_matrix); cerr << " )" << endl;
  cerr << "         * note    : parameter is a C-like matrix of 'A'x'B' cols x rows " << endl;
  cerr << "         * example : -M \"{{1,0,1},{0,1,1}}\"    (for A=3 and B=2) " << endl;
  cerr << "     b) SubsetSeed/VectorizedSubsetSeed "<< endl;
  cerr << "      -V                      : select the Vectorized subset seed model (default = disabled)"<< endl;
  cerr << "      -S {{<int>}} table      : vectorized subset seed Scoring table " << endl;
  cerr << "      -T <int>                : vectorized subset seed scoring Threshold " << endl;
  cerr << "                                                                               " << endl;
  cerr << "   2) Alignments:" << endl;
  cerr << "      -b <double>,<double>,...: fix the Bernoulli/Markov background distribution (size |'A'|^k)" << endl;
  cerr << "      -f <double>,<double>,...: fix the Bernoulli/Markov foreground distribution (size |'A'|^k)" << endl;
  cerr << "      -l <int>                : fix the alignment length (default = " <<  gv_alignment_length << ") " << endl;
  cerr << "      -u <int>,<int>,...      : set the homogeneous alignment scoring system (default = disabled)" << endl;
  cerr << "      -x <int>                : set the homogeneous alignment scoring system score (default score = " << gv_homogeneous_score <<")" << endl;
  cerr << "                                                                               " << endl;	
  cerr << "   3) Seed Enumeration:" << endl;
  cerr << "      -s <int>,<int>          : seed span (default min = " << gv_minspan << ",max = " << gv_maxspan << ")" << endl;
  cerr << "      -n <int>                : number of consecutive seeds to design  (default = " << gv_nbseeds << ", max = " << MAXSEEDS <<")" << endl;
  cerr << "      -r <int>                : random enumeration  (default = " << gv_nbruns << " is complete enumeration)" << endl;
  cerr << "      -k                      : hill climbing heuristic (check all permutations of 2 elements inside one seed)" << endl;
  cerr << "      -a <double>             : hill climbing heuristic frequency (good seeds are more likely to be optimized)" << endl;
  cerr << "      -w <double>,<double>    : minimal/maximal weight (default min = " << gv_minweight << ", max = " << gv_maxweight << ")" << endl;
  cerr << "         * symbol # (if defined, otherwise the max of all symbols) is of weight 1.0 " << endl;
  cerr << "         * symbol _ is of weight 0.0 " << endl;
  cerr << "         * symbol e weight is given by 'log_#(e matches background distribution)'" << endl;
  cerr << "      -i <int>,<int>,...      : signature id (number of '0','1',...'B-1' elements inside a subset seed)" << endl;
  cerr << "      -j <double>             : difference of weight (given as a ratio) allowed between consecutive seeds (value is between 0 and 1, default = " << gv_jive << ")" << endl;  
  cerr << "      -c <int>                : consider sensitivity when the database is only indexed on 1/c of its positions (default c = " << gv_cycle << " )" << endl;
  cerr << "      -g                      : consider sensitivity when the database is only indexed on 1/c of its positions and phasing is fixed" << endl;
  cerr << "      -d                      : disable Hopcroft minimization (default is enabled) " << endl;
  cerr << "      -m <seeds>              : check the given seed selectivity/sensitivity " << endl;
  cerr << "                                                                               " << endl;
  cerr << "   4) Miscellaneous:" << endl;
  cerr << "      -BSymbols '<char>'      : chain of seed alphabet symbols (default is '01..9AB..Zab..z')" << endl;
  cerr << "      -e <filename>           : input an initial \"pareto set\" file before optimizing" << endl;
  cerr << "      -o <filename>           : output the \"pareto set\" to this file after optimizing" << endl;
  cerr << "      -t                      : sleep during working hours [8h to 20h from monday to friday]" << endl;
  cerr << "                                                                               " << endl;
  cerr << "   5) Nucleic spaced seeds: (shortcuts that simulates hedera parameters)" << endl;
  cerr << "      -transitive             : \"-A 3 -B 3 -f 0.15,0.15,7 -b 0.5,0.25,0.25 -BSymbols '-@#' -l 64 -w 9,9 -s 10,15 -i 0,2,8\"" << endl;
  cerr << "      -spaced                 : \"-A 2 -B 2 -f 0.3,0.7     -b 0.75,0.25     -BSymbols '-#'  -l 64 -w 9,9 -s 9,15\"" << endl;
  exit(-1);
}


void DISPLAYMATRIX( std::ostream & os, vector< vector<int> > & matrix ) {
  os << "{";
  for (int b=0; b<gv_seed_alphabet_size; b++) {
    os << "{";
    for (int a=0; a<gv_align_alphabet_size; a++) {
      os << (matrix[a][b]);
      if (a < (gv_align_alphabet_size-1)) os << ",";
    }
    os << "}";
    if (b < (gv_seed_alphabet_size-1)) os << ",";
  }
  os << "}";
}

void DISPLAYTABLE( std::ostream & os, vector<double> & table ) {
  os << "{";
  for (int a=0; a<(int)table.size(); a++) {
    os << (table[a]);
    if (a < (int)(table.size()-1))
      os << ",";
  }
  os << "}";
}


void CHECKPROB(vector<double> & table) {
  double result = 0.0;
  for(int j=0 ; j<(int)table.size() ; j++)
    result += (table)[j];
  if(result < 0.99 || result > 1.01)
    ERROR("CHECKPROB"," sum of " << (table)[0] << " + ... + " << (table)[table.size()-1] << " must be 1.0 and not " << result);
}

void VERS() {
  cerr << PACKAGE << " version " << VERSION << endl;
  exit(-1);
}

void PARSEFLAG(int & i, char ** argv, int argc, bool & var, bool value) {
  var = value;
}



void PARSEINT(int & i,  char ** argv, int argc, int & var, int min, int max) {
  i++;
  if (i >= argc)
    ERROR("PARSEINT","\"" << argv[i-1] << "\" found without argument");
  char * next = argv[i];
  int i_tmp = strtol(argv[i],&next,10);
  if (errno != 0 || ((*next) != '\0'))
    ERROR("PARSEINT","\"" << argv[i-1] << "\" is not followed by an integer");
  if (i_tmp>max || i_tmp <min)
    ERROR("PARSEINT","\"" << argv[i-1] << "\" is followed by an integer outside the valid range");
  var = i_tmp;
}


void PARSEDINT(int & i, char ** argv, int argc,
	       int & var1, int min1, int max1,
	       int & var2, int min2, int max2) {
  i++;
  if (i >= argc)
    ERROR("PARSEDINT","\"" << argv[i-1] << "\" found without argument");
  int i_tmp = sscanf(argv[i],"%i,%i",&var1,&var2);
  if (i_tmp != 2 )
    ERROR("PARSEDINT","\"" << argv[i-1] << "\" is not followed by an <int>,<int>");
  if (var1 < min1 || var1 > max1 || var2 < min2 || var2 > max2)
    ERROR("PARSEDINT","\"" << argv[i-1] << "\" is followed by integers out of the range");
  if (var1 > var2)
    ERROR("PARSEDINT","\"" << var1 << "\" must be <= \"" << var2 << "");
}



void PARSEDOUBLE(int & i, char ** argv, int argc, double & var , double min, double max) {
  i++;
  if (i >= argc)
    ERROR("PARSEDOUBLE","\"" << argv[i-1] << "\" found without argument");
  char * next = argv[i];
  double d_tmp = strtod(argv[i],&next);
  if (errno != 0 || ((*next) != '\0'))
    ERROR("PARSEDOUBLE","\"" << argv[i-1] << "\" is not followed by a double") ;
  if (d_tmp>max || d_tmp <min)
    ERROR("PARSE","\"" << argv[i-1] << "\" is followed by a double outside the valid range");
  var = d_tmp;
}


void PARSEDDOUBLE(int & i, char ** argv, int argc,
		  double & var1, double min1, double max1,
		  double & var2, double min2, double max2) {
  i++;
  if (i >= argc)
    ERROR("PARSEDDOUBLE","\"" << argv[i-1] << "\" found without argument");
  int i_tmp = sscanf(argv[i],"%lf,%lf",&var1,&var2);
  if (i_tmp != 2 )
    ERROR("PARSEDDOUBLE","\"" << argv[i-1] << "\" is not followed by an <dbl>,<dbl>");
  if (var1 < min1 || var1 > max1 || var2 < min2 || var2 > max2)
    ERROR("PARSEDDOUBLE","\"" << argv[i-1] << "\" is followed by doubles out of the range");
  if (var1 > var2)
    ERROR("PARSEDDOUBLE","\"" << var1 << "\" must be <= \"" << var2 << "");
}



void PARSESIGNATURE(int & i, char ** argv, int argc, vector<int> & table) {
  i++;
  if (i >= argc)
    ERROR("PARSESIGNATURE","\"" << argv[i-1] << "\" found without argument");

  /* read table */
  char * pch = strtok(argv[i],",;");
  table.clear();
  while (pch){
    int value = 0;
    int i_tmp = sscanf(pch,"%d",&value);
    if (i_tmp != 1)
      ERROR("PARSESIGNATURE","\"" << pch << "\" is not a correct integer");
    if (value < 0 || value >= 64)
      ERROR("PARSESIGNATURE","\"" << pch << "\" is out of the range [0..N] (N=64)");
    table.push_back(value);
    pch = strtok(NULL,",;");
  }

  /* check table size */
  if ((int)table.size() != gv_seed_alphabet_size)
    ERROR("PARSESIGNATURE", " there is not a correct number of <int> values given by " << argv[i-1] << " when compared with the current seed alphabet size B = " << gv_seed_alphabet_size);
}


void PARSEHOMOGENEOUS(int & i, char ** argv, int argc, vector<int> & table) {
  i++;
  if (i >= argc)
    ERROR("PARSEHOMOGENEOUS","\"" << argv[i-1] << "\" found without argument");

  /* read table */
  char * pch = strtok(argv[i],",;");
  table.clear();
  while (pch){
    int value = 0;
    int i_tmp = sscanf(pch,"%d",&value);
    if (i_tmp != 1)
      ERROR("PARSEHOMOGENEOUS","\"" << pch << "\" is not a correct integer");
    if (value < -1000 || value > 1000)
      ERROR("PARSEHOMOGENEOUS","\"" << pch << "\" is out of the range [-1000..1000]");
    table.push_back(value);
    pch = strtok(NULL,",;");
  }

  /* check table size */
  if ((int)table.size() != gv_align_alphabet_size)
    ERROR("PARSEHOMOGENEOUS", " there is not a correct number of <int> values given by " << argv[i-1] << " when compared with the current align alphabet size A = " << gv_align_alphabet_size);
}


void PARSEPROBS(int & i, char ** argv, int argc,
		vector<double> & table, int & k) {
  i++;
  if (i >= argc)
    ERROR("PARSEPROBS","\"" << argv[i-1] << "\" found without argument");

  /* read table */
  char * pch = strtok(argv[i],",;");
  table.clear();
  while (pch){
    double value = 0.0;
    int i_tmp = sscanf(pch,"%lf",&value);
    if (i_tmp != 1)
      ERROR("PARSEPROBS","\"" << pch << "\" is not a correct <dbl>");
    if (value < 0 || value > 1 )
      ERROR("PARSEPROBS","\"" << pch << "\" is  out of the range [0,1]");
    table.push_back(value);
    pch = strtok(NULL,",;");
  }

  /* check table size */
  int a = gv_align_alphabet_size;
  k = 0;
  while(1) {
    if ((int)table.size() == a)
      break;
    else if ((int)table.size() < a)
      ERROR("PARSEPROBS", " there is not a correct number of <dbl> values ("<< (table.size()) <<") given by " << argv[i-1] << " when compared with the current alphabet size A = " << gv_align_alphabet_size);
    a *= gv_align_alphabet_size;
    k++;
  }
  CHECKPROB(table);
}







typedef enum {
  dummy,
  readfirstopenbracket,
  readsecondopenbracket,
  readint,
  readintseparator,
  readsecondclosedbracket,
  readsecondbracketseparator,
  readfirstclosedbracket,
  readfirstbracketseparator
} automatareadmatrixstates;




void PARSEMATRIX(int & i, char ** argv, int argc, vector< vector<int> > & matrix, int min, int max, int gnbcolumns, int gnbrows) {

  i++;
  if (i >= argc)
    ERROR("PARSEMATRIX","\"" << argv[i-1] << "\" found without argument");
  char * pch = argv[i];
  int nbcolumns = 0, nbrows = 0;

  automatareadmatrixstates state = dummy;
  /* automata */
  while(*pch) {

    switch (*pch){

    case '{':
      switch (state){
      case dummy:
	nbrows = 0;
	state = readfirstopenbracket;
	break;
      case readfirstopenbracket:
      case readsecondclosedbracket:
      case readsecondbracketseparator:
	nbcolumns = 0;
	nbrows++;
	if (nbrows > gnbrows)
	  ERROR("PARSEMATRIX","\" invalid number of lines : " << nbrows << " != " << gnbrows << " in \"" << argv[i] << "\" string parameter ");

	state = readsecondopenbracket;
	break;
      default:
	ERROR("PARSEMATRIX","\" invalid \'" << *pch << "\' in \"" << argv[i] << "\" string parameter ");
      }
      pch++;
      break;

    case '}':
      switch (state){
      case readint:
      case readintseparator:
	state = readsecondclosedbracket;
	if (nbcolumns != gnbcolumns)
	  ERROR("PARSEMATRIX","\" invalid number of columns :  " << nbcolumns << " != " << gnbcolumns << " in \"" << argv[i] << "\" string parameter ");
	break;
      case readsecondclosedbracket:
	state = readfirstclosedbracket;
	if (nbrows != gnbrows )
	  ERROR("PARSEMATRIX","\" invalid number of lines : " << nbrows << " != " << gnbrows << " in \"" << argv[i] << "\" string parameter ");
	break;
      default:
	ERROR("PARSEMATRIX","\" invalid \'" << *pch << "\' in \"" << argv[i] << "\" string parameter ");
      }
      pch++;
      break;

    case ' ':
    case '\t':
      pch++;
      break;

    case ';':
    case ',':
      switch (state){
      case readint:
	state = readintseparator;
	break;
      case readsecondclosedbracket:
	state = readsecondbracketseparator;
	break;
      case readfirstclosedbracket:
	state = readfirstbracketseparator;
	break;
      default:
	ERROR("PARSEMATRIX","\" invalid \'" << *pch << "\' in \"" << argv[i] << "\" string parameter ");
      }
      pch++;
      break;

    default: /* read value */
      if (state == readsecondopenbracket || state == readintseparator) {
	state = readint;
	int value = strtol (pch, &pch, 10);
	if (value>max || value<min)
	  ERROR("PARSEMATRIX","\"" << argv[i] << "\" contains at least one integer outside the valid range");
	if (nbcolumns >= gnbcolumns)
	  ERROR("PARSEMATRIX","\" invalid number of columns : " << (nbcolumns+1) << ">" <<  gnbcolumns << " in \"" << argv[i] << "\" string parameter ");
	assert(nbrows-1<gnbrows);
	matrix[nbcolumns][nbrows-1] = value;
	nbcolumns++;

      } else {
	ERROR("PARSEMATRIX","\" invalid \'" << *pch << "\' in \"" << argv[i] << "\" string parameter ");
      }
    }
  } /* while */
  if (state == readfirstclosedbracket || state == readfirstbracketseparator)
    return;
  else
    ERROR("PARSEMATRIX","\" unfinished string parameter \"" << argv[i] << "\" ");

}



void CHECKMATCHINGMATRIX( vector< vector<int> > & matrix ) {
  gv_matching_symbol_flag = true;

  /* check the '1' symbol existance */ 
  for (int b=0; b<gv_seed_alphabet_size; b++) {
    if ( matrix[gv_align_alphabet_size-1][b] == 0 ) {
      gv_matching_symbol_flag = false;
    }
  }
  /* check the '#' symbol existance */
  for(int a=0; a <gv_align_alphabet_size-1;a++) {
    if ( matrix[a][gv_seed_alphabet_size-1] == 1 ) {
      gv_matching_symbol_flag = false;
    }
  }
  if ( matrix[gv_align_alphabet_size-1][gv_seed_alphabet_size-1] == 0 ) {
    gv_matching_symbol_flag = false;
  }
}


void PARSESYMBOLS(int & i, char ** argv, int argc, int size, char * &array) {
  i++;
  if (i >= argc)
    ERROR("PARSESYMBOLS","\"" << argv[i-1] << "\" found without argument");  
  if (strlen(argv[i]) != (unsigned) size)
    ERROR("PARSESYMBOLS","\"" << argv[i] << "\" is not of required length " << size);  
  array = new char[size];
  for(char * d = array, *s = argv[i]; *s ; d++, s++){
    *d = *s;
    cerr << "\'" << (*d) << "\'" << endl;
  }
}

void PARSESEEDS(int & i, char ** argv, int argc) {
  i++;
  if (i >= argc)
    ERROR("PARSESEEDS","\"" << argv[i-1] << "\" found without argument");

  gv_motif   = 1;
  gv_nbseeds = 0;

  char * pch = strtok(argv[i],",;");
  while (pch) {
    string s = string(strdup(pch));
    gv_seeds[gv_nbseeds] = new seed(s);
    gv_nbseeds++;
    if ( gv_nbseeds > MAXSEEDS)
      ERROR("PARSESEEDS","\" element " << gv_nbseeds << "\" is out of array bounds");
    pch = strtok(NULL,",;");
  }
}


void PARSEINPUT(int & i, char ** argv, int argc) {
  i++;
  if (i >= argc)
    ERROR("PARSEINPUT","\"" << argv[i-1] << "\" found without argument");
  
  char * pch = strtok(argv[i],",;");
  while (pch) {
    char * file = strdup(pch);
    gv_input_filenames[gv_nb_input_filenames++] = file;
    pch = strtok(NULL,",;");
  }
}

void PARSEOUTPUT(int & i, char ** argv, int argc) {
  i++;
  if (i >= argc)
    ERROR("PARSEINPUT","\"" << argv[i-1] << "\" found without argument");
  gv_output_filename = strdup(argv[i]);
}


void SCANARG(int argc , char ** argv) {
  for(int i=1; i<argc; i++){

    /* 1) Seed model */
    if(!strcmp(argv[i],"-A")||!strcmp(argv[i],"--A")) {
      PARSEINT(i,argv,argc,gv_align_alphabet_size,2,1024);
      build_default_subsetseed_matching_matrix();
      build_default_vectorizedsubsetseed_scoring_matrix();
      build_default_probabilities();
      gv_homogeneous_flag = false;
    } else if(!strcmp(argv[i],"-B")||!strcmp(argv[i],"--B")) {
      PARSEINT(i,argv,argc,gv_seed_alphabet_size,1,1024);
      build_default_subsetseed_matching_matrix();
      build_default_vectorizedsubsetseed_scoring_matrix();
      build_default_probabilities();
      gv_signature_flag = false;
      gv_bsymbols_flag  = false;
    } else if(!strcmp(argv[i],"-M")||!strcmp(argv[i],"--Matching")) {
      PARSEMATRIX(i,argv,argc,gv_subsetseed_matching_matrix,0,1,gv_align_alphabet_size,gv_seed_alphabet_size);
      CHECKMATCHINGMATRIX(gv_subsetseed_matching_matrix);
    } else if(!strcmp(argv[i],"-S")||!strcmp(argv[i],"--Scoring")) {
      PARSEMATRIX(i,argv,argc,gv_vectorizedsubsetseed_scoring_matrix,-10000,10000,gv_align_alphabet_size,gv_seed_alphabet_size);
    } else if(!strcmp(argv[i],"-T")||!strcmp(argv[i],"--Threshold")) {
      PARSEINT(i,argv,argc,gv_vectorizedsubsetseed_scoring_threshold,-10000,+10000);
    } else if(!strcmp(argv[i],"-V")||!strcmp(argv[i],"--Vectorized")) {
      PARSEFLAG(i,argv,argc,gv_vectorized_flag,true);
    

    /* 2) Alignment model */
    } else if(!strcmp(argv[i],"-b")||!strcmp(argv[i],"--background")) {
      PARSEPROBS(i,argv,argc,gv_bsel,gv_bsel_k);
    } else if(!strcmp(argv[i],"-f")||!strcmp(argv[i],"--foreground")) {
      PARSEPROBS(i,argv,argc,gv_bsens,gv_bsens_k);
    } else if(!strcmp(argv[i],"-l")||!strcmp(argv[i],"--length")) {
      PARSEINT(i,argv,argc,gv_alignment_length,1,1000000);
    } else if(!strcmp(argv[i],"-u")||!strcmp(argv[i],"--homogeneous")) {
      PARSEHOMOGENEOUS(i,argv,argc,gv_homogeneous_scores);
      gv_homogeneous_flag = true;
    } else if(!strcmp(argv[i],"-x")||!strcmp(argv[i],"--homogeneous-threshold")) {
      PARSEINT(i,argv,argc,gv_homogeneous_score,1,1000000);
    /* 3) Seed enumeration */
    } else if(!strcmp(argv[i],"-s")||!strcmp(argv[i],"--span")) {
      PARSEDINT(i,argv,argc,gv_minspan,1,1000,gv_maxspan,1,1000);      
    } else if(!strcmp(argv[i],"-w")||!strcmp(argv[i],"--weight")) {
      gv_weight_interval = 1;
      PARSEDDOUBLE(i,argv,argc,gv_minweight,0,1e32,gv_maxweight,0,1e32);
    } else if(!strcmp(argv[i],"-n")||!strcmp(argv[i],"--number")) {
      PARSEINT(i,argv,argc,gv_nbseeds,1,MAXSEEDS);
    } else if(!strcmp(argv[i],"-r")||!strcmp(argv[i],"--random")) {
      PARSEINT(i,argv,argc,gv_nbruns,1,1000000000);
    } else if(!strcmp(argv[i],"-k")||!strcmp(argv[i],"--hillclimbing")) {
      PARSEFLAG(i,argv,argc,gv_hillclimbing_flag,true);
    } else if(!strcmp(argv[i],"-a")||!strcmp(argv[i],"--hillclimbing_alpha")) {
      PARSEDOUBLE(i,argv,argc,gv_hillclimbing_alpha,0,1);
    } else if(!strcmp(argv[i],"-d")||!strcmp(argv[i],"--disable")) {
      PARSEFLAG(i,argv,argc,gv_minimize_flag,false);
    } else if(!strcmp(argv[i],"-m")||!strcmp(argv[i],"--motifs")) {
      PARSESEEDS(i,argv,argc);
    } else if(!strcmp(argv[i],"-i")||!strcmp(argv[i],"--idsign")) {
      PARSESIGNATURE(i,argv,argc,gv_signature);
      gv_signature_flag = true;
    } else if(!strcmp(argv[i],"-c")||!strcmp(argv[i],"--cycle")) {
      PARSEINT(i,argv,argc,gv_cycle,1,1000);
      gv_cycle_flag = true;
    } else if(!strcmp(argv[i],"-g")||!strcmp(argv[i],"--fixedphase")) {
      PARSEFLAG(i,argv,argc,gv_cycle_fixedphase_flag,true);
      gv_cycle_flag = true;
    } else if(!strcmp(argv[i],"-j")||!strcmp(argv[i],"--jive")) {
      PARSEDOUBLE(i,argv,argc,gv_jive,0,1);
      /* 4) misc */
    } else if(!strcmp(argv[i],"-e")||!strcmp(argv[i],"--enter")) {
      PARSEINPUT(i,argv,argc);
    } else if(!strcmp(argv[i],"-o")||!strcmp(argv[i],"--output")) {
      PARSEOUTPUT(i,argv,argc);
    } else if(!strcmp(argv[i],"-t")||!strcmp(argv[i],"--time")) {
      PARSEFLAG(i,argv,argc,gv_sleeping_flag,true);
    } else if(!strcmp(argv[i],"-h")||!strcmp(argv[i],"--help")) {
      USAGE();
    } else if(!strcmp(argv[i],"-v")||!strcmp(argv[i],"--version")) {
      VERS();
      /* 5) */
    } else if (!strcmp(argv[i],"-BSymbols")) {
      PARSESYMBOLS(i,argv,argc,gv_seed_alphabet_size,gv_bsymbols_array);
      gv_bsymbols_flag = true;      
    } else if (!strcmp(argv[i],"-transitive")) {
      /* transitives params */
      gv_align_alphabet_size = 3;
      gv_seed_alphabet_size  = 3;
      gv_signature_flag      = true;
      gv_signature = vector<int>(3); gv_signature[0] = 0; gv_signature[1] = 2; gv_signature[2] = 8;
      gv_bsens = vector<double>(3); gv_bsens[0] = 0.15; gv_bsens[1] = 0.15; gv_bsens[2] = 0.7;
      gv_bsens_k = 0;
      gv_bsel  = vector<double>(3); gv_bsel[0]  = 0.5;  gv_bsel[1]  = 0.25; gv_bsel[2]  = 0.25;
      gv_bsel_k = 0;
      gv_bsel_weight = vector<double>(3); gv_bsel_weight[0] = 0.0; gv_bsel_weight[1] = 0.5; gv_bsel_weight[2] = 1.0;
      gv_bsel_minprob = 0.25;
      gv_bsel_maxprob = 1.0;
      gv_minspan = 10;
      gv_maxspan = 15;
      gv_minweight = 9; gv_maxweight = 9;
      gv_weight_interval = true;
      gv_vectorized_flag = false;
      gv_bsymbols_array = "-@#";
      gv_bsymbols_flag = true;
      build_default_subsetseed_matching_matrix();
      gv_alignment_length = 64;
    } else if (!strcmp(argv[i],"-spaced")) {
      /* spaced params */
      gv_align_alphabet_size = 2;
      gv_seed_alphabet_size  = 2;
      gv_signature_flag      = false;
      gv_bsens = vector<double>(2); gv_bsens[0] = 0.3; gv_bsens[1] = 0.7;
      gv_bsens_k = 0;
      gv_bsel  = vector<double>(2); gv_bsel[0]  = 0.75;  gv_bsel[1]  = 0.25;
      gv_bsel_k = 0;
      gv_bsel_weight = vector<double>(2); gv_bsel_weight[0] = 0.0; gv_bsel_weight[1] = 1.0;
      gv_bsel_minprob = 0.25;
      gv_bsel_maxprob = 1.0;
      gv_minspan = 9;
      gv_maxspan = 15;
      gv_minweight = 9; gv_maxweight = 9;
      gv_weight_interval = true;
      gv_vectorized_flag = false;
      gv_bsymbols_array = "-#";
      gv_bsymbols_flag  = true;
      build_default_subsetseed_matching_matrix();
      gv_alignment_length = 64;
    } else ERROR("PARSE","\"" << argv[i] << "\" is not a valid argument");
  }
}













/**
 * @class seedproperties
 * @brief simple "inner class" to keep seed properties as
 * sensitivity, selectivity ... and other properties
 */

class seedproperties {

 public:

  seedproperties(double sel, double sens, double dist, std::string str){
    this->sel   = sel;
    this->sens  = sens;
    this->dist  = dist;
    this->str   = string(str);
  }

  double sel;
  double sens;
  double dist;
  string str;
  int    phase;

  /**
   * @brief comparison operator (sort on selectivity first, then on sensitivity for equal selectivity)
   */

  bool operator<(seedproperties & e) {
    return (this->sel < e.sel) || ( (this->sel == e.sel) && (this->sens < e.sens)) ;
  }


  friend std::ostream& operator<<(std::ostream& os, seedproperties& e);

};



std::ostream& operator<<(std::ostream& os, seedproperties& e){
  os <<  e.str << "\t" << e.sel <<  "\t" << e.sens <<  "\t" << e.dist ;
  return os;
}



double dist(double sel,double sens) {
  return sqrt((1.0 - sel)*(1.0 - sel) + (1.0 - sens)*(1.0 - sens));
}


/**
 * @brief display a list of seeds properties
 */

int display (list<seedproperties> & l) {
  cerr << "#" << "\t" << "1th column : seed motif"
              << "\t" << "2nd column : selectivity"
              << "\t" << "3rd column : sensitivity"
              << "\t" << "4th column : distance to (1,1)" ;

  l.sort();
  for(list<seedproperties>::iterator i = l.begin(); i != l.end() ; i++) {
    cout << (*i) << endl;
  }
  return 0;

}


/**
 * @brief Compute the convex hull area for the selectivity/sensitivity of a set of seeds
 * @brief [this definition is not suitable and must NOT be used]
 */

double areaConvex(list<seedproperties> & l) {


  /* (1) push border seedproperties and sort */
  l.sort();
  l.push_front(seedproperties(0.0,1.0,1.0,string("_")));
  l.push_back( seedproperties(1.0,0.0,1.0,string("_")));

  /* (2) compute convex hull */
  list<seedproperties>::iterator i = l.begin();
  vector<seedproperties>         c = vector<seedproperties>();
  vector<double>          s = vector<double>();
  c.push_back(*(i));
  s.push_back(+0.0);
  i++;

  while(i != l.end()) {
    /* some elments can be negative */
    if ((i->sel >= 0) && (i->sens >= 0)) {
    eval:
      /* increased x-sel */
      if (i->sel > c.back().sel){
	double slope = (i->sens - c.back().sens)/(i->sel - c.back().sel);
	
	if (slope <= s.back()){
	  c.push_back(*i);
	  s.push_back(slope);
	}else{
	  if (s.size() > 1){
	    c.pop_back();
	    s.pop_back();
	    goto eval;
	  }
	}
	
      }else{
	
	/* same x-sel but increased sens */
	if ( i->sens > c.back().sens){
	  c.pop_back();
	  s.pop_back();
	  double slope = (i->sens - c.back().sens)/(i->sel - c.back().sel);
	
	  if (slope <= s.back()){
	    c.push_back(*i);
	    s.push_back(slope);
	  }else{
	    if (s.size() > 1){
	      c.pop_back();
	      s.pop_back();
	      goto eval;
	    }
	  }
	
	}
      }
    }
    i++;
  }

  /* (3) compute covered area */
  double area = 0.0;
  vector<seedproperties>::iterator j = c.begin();
  vector<double>::iterator  k = s.begin();
  for( vector<seedproperties>::iterator i = c.begin() ; i !=  c.end() ; i++ ){
    area += 0.5 * (i->sel - j->sel) * (i->sens + j->sens) ;
    j = i;
    //cout << (*i) << endl;
  }
  cerr << "convex hull area : " << area << endl;
  return area;
}




/**
 * @brief Select seeds that are on the current "pareto set"
 * @brief usefull to keep best seeds only
 * @return the pareto area as the sum of rectangles covered by each sel/sens coordinate
 */


double selectPareto(list<seedproperties> & l) {

  /* (1) push border seedproperties and sort */
  l.sort();

  /* (2) compute border hull */
  list<seedproperties>::iterator i = l.begin();
  list<seedproperties>::iterator j = l.begin();
  i++;

  while(i != l.end()) {

    /* increased sensitivity */
    if (i->sens > j->sens /* note that i->sel >= j-> sel */) {
      if (j != l.begin()) {
	list<seedproperties>::iterator k = j;
	j--;
	l.erase(k);
      } else {
	l.erase(j);
	j = i;
	i++;
      }
    } else {
      /* increased selectivity */
      if (i->sel > j->sel) {
	j = i;
	i++;
      } else {
	/* equal selectivity but decreased sensitivity */
	list<seedproperties>::iterator k = i;
	i++;
	l.erase(k);
      }
    }
  } /* while */



  /* (3) compute covered area */
  double area = 0.0;
  j = l.begin();
  for( i = l.begin() ; i !=  l.end() ; i++ ){
    area += (i->sel - j->sel) * (i->sens) ;
    j = i;
    //cerr << (*i) << endl;
  }
  return area;
}



/**
 * @brief Compute the "pareto set" area for the selectivity/sensitivity of a set of seeds
 * @brief [this definition is more suitable]
 * @return the pareto area as the sum of rectangles covered by each sel/sens coordinate
 */

double list_and_areaPareto(list<seedproperties> & l) {


  /* (1) select pareto set and compute area */
  double area = selectPareto(l);
  cerr << "pareto area : " << area << endl;
  /* (2) display seeds */
  for( list<seedproperties>::iterator i = l.begin() ; i !=  l.end() ; i++ ){
    cout << (*i) << endl;
  }
  return area;
}


/**
 * @brief this method check if a given sel/sens is out of the pareto curve and thus does improve the overall selectivity/sensitivity 
 * @return true if the element has been insered
 */

double insert_Pareto(list<seedproperties> & l, seedproperties & e) {
  /* seach the position where to insert */
  list<seedproperties>::iterator i = l.begin();
  list<seedproperties>::iterator j = l.begin();  
  while(i != l.end() && e.sel > i->sel) {
    j = i;
    i++;
  }
  
  /* insert (front/back/between) */
  if ( i ==  l.end()) {
    l.push_back(e);
    return ABS(j->sens - e.sens)/(1.0 - e.sens);;
  } else if ( i == l.begin() ) {
    l.push_front(e);
    return ABS(i->sens - e.sens)/(1.0 - e.sens);;
  } else {
    /*  e.sel <= i->sel */
    if (i->sel == e.sel) {
      double dist = e.sens - i->sens;
      if (dist > 0) {
	/* replace i by e */
	l.insert(i,e);
	l.erase(i);
	return dist/(1.0 - e.sens);
      } else {
	return dist/(1.0 - e.sens);
      }
    } else { 
      double dist = e.sens - i->sens;
      /*  j-> sel < e.sel < i->sel */
      if (dist > 0) {
	/* add e */
	l.insert(i,e);
	return dist/(1.0 - e.sens);;
      } else {
	return dist/(1.0 - e.sens);;
      }
    }
  }    
}






/*
 * @brief read a list of initial seeds and merge them in the pareto set 
 * @param l gives the list to complete the pareto part
 * @param filename gives the input file name 
 * @return error code (always 0, even if the file is unreadable)
 */

int inputPareto(list<seedproperties> & l, char * filename) {
  ifstream in;
  in.open(filename, ifstream::in);
  while(in.good() && !in.eof()) {
    string str;
    double sel = 0.0, sens = 0.0, dist = 0.0;
    in >> str >> sel >> sens >> dist ;    
    l.push_back(seedproperties(sel,sens,dist,str));
  }  
  in.close();
  selectPareto(l);
  return 0;
}
 


/*
 * @brief write the pareto set to a file
 * @param l gives the list to write
 * @param filename gives the output file name
 * @return error code (always 0, but quit if the file in unwritable)
 */

int outputPareto(list<seedproperties> & l, char * filename) {
  // select pareto
  selectPareto(l);
  // write pareto
  ofstream out;
  out.open(filename, ifstream::out);
  if (out.good()) {
    for( list<seedproperties>::iterator i = l.begin() ; i !=  l.end() ; i++ ){
      out << (*i) << endl;
    }
  } else {
    ERROR("outputPareto"," unable to write into \"" << filename << "\"");
  }
  out.close();
  return 0;
}
  
 







/*
 *@ brief this method is needeed when |A| or |B| size has changed
 */

int build_default_subsetseed_matching_matrix() {
  /* build matrix */
  gv_subsetseed_matching_matrix.clear();
  for (int a = 0 ; a < gv_align_alphabet_size ; a++) {
    vector<int> * v = new vector<int>(gv_seed_alphabet_size,0);
    gv_subsetseed_matching_matrix.push_back(*v);
  }

  /* fill it in */
  for(int b = 0 ; b < gv_seed_alphabet_size ; b++ ) {
    /* sharp */
    gv_subsetseed_matching_matrix[gv_align_alphabet_size-1][b] = 1;
    for(int a = b; a < gv_align_alphabet_size ; a++ )
      gv_subsetseed_matching_matrix[a][b] = 1;
  }

  /* set the flag to signal the '#' element */
  CHECKMATCHINGMATRIX(gv_subsetseed_matching_matrix);
  return 0;
}




/**
 *@ brief this method is needeed when |A| or |B| size has changed
 */

int build_default_vectorizedsubsetseed_scoring_matrix() {

  /* build matrix */
  gv_vectorizedsubsetseed_scoring_matrix.clear();
  for (int a = 0 ; a < gv_align_alphabet_size ; a++) {
    vector<int> * v = new vector<int>(gv_seed_alphabet_size,-1);
    gv_vectorizedsubsetseed_scoring_matrix.push_back(*v);
  }

  /* fill it in */
  for (int a = 0 ; a < gv_align_alphabet_size ; a++) {
    for(int b = 0 ; b < gv_seed_alphabet_size ; b++ ) {
      if (gv_subsetseed_matching_matrix[a][b])
	gv_vectorizedsubsetseed_scoring_matrix[a][b] = 1;
    }
  }
  return 0;
}



/**
 *@ brief thie method is needeed when |A| size has changed
 */

int build_default_probabilities()  {
  gv_bsens          = vector <double>(gv_align_alphabet_size,0);
  gv_bsens_k        = 0;
  gv_bsel           = vector <double>(gv_align_alphabet_size,0);
  gv_bsel_k         = 0;

  double remaining  = 1.0;

  /* bsel : 0.5,0.25,0.125,0.125,
   * bsens : 0.125,0.125,0.25,0.5
   */

  for(int i = 0 ; i < gv_align_alphabet_size; i++) {
    if ( i < (gv_align_alphabet_size - 1) ) {
      remaining *= .5;
    }

    gv_bsel[i] = remaining;
    gv_bsens[gv_align_alphabet_size-i-1] = remaining;
  }

  return 0;
}

void computeWeight() {
  gv_bsel_weight  = vector<double>(gv_seed_alphabet_size);
  gv_bsel_minprob =  1e32;
  gv_bsel_maxprob = -1e32;

  /* compute background probs */
  for(int b = 0; b < gv_seed_alphabet_size ; b++) {
    double pr = 0.0;
    for(int a = 0; a < gv_align_alphabet_size ; a++) {
      if (gv_subsetseed_matching_matrix[a][b]) {
	for (int i = 0; i < power(gv_align_alphabet_size,gv_bsel_k) ; i++)
	  pr += gv_bsel[a+i*gv_align_alphabet_size];
      }
    }
    gv_bsel_weight[b] = pr;
    gv_bsel_minprob   = MIN(pr,gv_bsel_minprob);
    gv_bsel_maxprob   = MAX(pr,gv_bsel_maxprob);
     
  }
  
  /* take the log_#(b) */
  for(int b = 0; b < gv_seed_alphabet_size ; b++) {
    gv_bsel_weight[b] = log(gv_bsel_weight[b])/log(gv_bsel_minprob) ;
  }
  
}




list<seedproperties> l ;


#ifdef SIGNAL

/*
 * @brief this handler writes file on premature ending
 */

void termin_handler( int signal ) { 
  cerr <<"Signal handler..." << endl;
  
  /* security input */
  if(gv_nb_input_filenames > 0) {
    for(int i=0;i<gv_nb_input_filenames;i++)
      inputPareto(l,gv_input_filenames[i]);
  }

  /* final output */
  cerr << "# seeds\t sel\t sens\t dist"<<  endl;
  if (!gv_output_filename)
    list_and_areaPareto(l);
  else
    outputPareto(l,gv_output_filename);
  exit(0);
}

#endif




int main(int argc, char * argv[]) {

  int nbruns = 0;

  l = list<seedproperties>();
  l.push_front(seedproperties(0.0,1.0,1.0,string("_")));
  l.push_back( seedproperties(1.0,0.0,1.0,string("_")));


  build_default_subsetseed_matching_matrix();
  build_default_vectorizedsubsetseed_scoring_matrix();
  build_default_probabilities();
  
  SCANARG(argc,argv);
  
  /* randomize */
  srand(GETMSTIME()*GETPID());
  
  /* initail pareto set */
  if(gv_nb_input_filenames > 0) {
    for(int i=0;i<gv_nb_input_filenames;i++)
      inputPareto(l,gv_input_filenames[i]);
  }

#ifdef SIGNAL
  signal(SIGTERM, &termin_handler );
  signal(SIGSTOP, &termin_handler );
#endif


  /*
   * O) DISPLAY PARAMETERS
   */
  cerr << "* alignment alphabet size |A| : " << gv_align_alphabet_size << " , " << "seed alphabet size |B| : "  << gv_seed_alphabet_size << endl;
  cerr << "* subset seed matching matrix : "; DISPLAYMATRIX(cerr,gv_subsetseed_matching_matrix); cerr << endl;
  if (gv_vectorized_flag) {
    cerr << "* subset seed scoring matrix : "; DISPLAYMATRIX(cerr,gv_vectorizedsubsetseed_scoring_matrix); cerr << endl;
    cerr << "* subset seed scoring threshold : " << gv_vectorizedsubsetseed_scoring_threshold << endl;
  }

  /* compute the weight for each seed element */
  computeWeight();
  cerr << "* seed alphabet weights " ; DISPLAYTABLE(cerr,gv_bsel_weight); cerr <<endl;

  /* build the foreground/background models */
  automaton a_sel;
  if (gv_bsel_k == 0) {
    a_sel.Automaton_Bernoulli(gv_bsel);
    cerr << "* background Bernoulli model "; DISPLAYTABLE(cerr,gv_bsel); cerr << endl;
  } else {
    a_sel.Automaton_Markov(gv_bsel,gv_bsel_k);
    cerr << "* background Markov model "; DISPLAYTABLE(cerr,gv_bsel);
    cerr << " (order : " << (gv_bsel_k)<< ")" << endl;
  }

  automaton a_sens;
  if (gv_bsens_k == 0) {
    a_sens.Automaton_Bernoulli(gv_bsens);
    cerr << "* foreground Bernoulli model "; DISPLAYTABLE(cerr,gv_bsens); cerr << endl;
  } else {
    a_sens.Automaton_Markov(gv_bsens,gv_bsens_k);
    cerr << "* foreground Markov model "; DISPLAYTABLE(cerr,gv_bsens);
    cerr << " (order : " << (gv_bsens_k) << ")" << endl;
  }

  /*
   * build the homogeneous automaton if requested and 
   * compute the foreground probability of homogeneous paths 
   */

  double pr_div_homogeneous = 1.00;
  automaton * a_homogeneous = new automaton();  
  if (gv_homogeneous_flag) {
    a_homogeneous->Automaton_Homogeneous(gv_homogeneous_scores, gv_homogeneous_score, gv_alignment_length);
    if (gv_minimize_flag) {
      automaton * na = a_homogeneous->Hopcroft(); 
      delete a_homogeneous; 
      a_homogeneous = na ;
    }
    automaton * a_product_homogeneous_sens = a_homogeneous->product(a_sens, false, true);
    pr_div_homogeneous = a_product_homogeneous_sens->PrFinal(gv_alignment_length);
  }



  /*
   * I) MAIN STATISTICAL PART
   */

  if ( gv_motif ) {
    
    /*
     * A) check a command line given set of seeds ( -m 1022,20102)
     */
    
    automaton * a_spr = NULL;
    double      selp = 0.0;

    for(int i=0 ; i < gv_nbseeds ; i++){

      /* seed automaton */
      
      automaton * a_s = new automaton();
	
      cerr << " seed : " << (*gv_seeds[i]) << endl;
      SEEDAUTOMATON(a_s,gv_seeds[i],false);
      cerr << " size of " << (*gv_seeds[i]) << " : " << (a_s->size()) << endl;;
      if (gv_minimize_flag) {
	automaton * na = a_s->Hopcroft(); 
	delete a_s; 
	a_s = na ;
      }
      cerr << " reduced size : " << (a_s->size()) << endl;;
      
      /* (1) compute partial selectivity */
      automaton * a_pr_s_sel      = a_s->product(a_sel, false, true);

      //FIXME  selp                        =  MAX(selp,a_pr_s_sel->PrFinal(gv_seeds[i]->span()));
      selp                        +=  a_pr_s_sel->PrFinal(gv_seeds[i]->span()); 
      
      delete a_pr_s_sel;


      /* (2) product automaton for sensitivity */

      if (i > 0) {
	automaton * a_spr_temp = a_spr;
	a_spr =  a_spr->product(*a_s, false, false);
	cerr << " product size : " << (a_spr->size()) << endl;;
      	if (gv_minimize_flag) {
	  automaton * na = a_spr->Hopcroft(); 
	  delete a_spr; 
	  a_spr = na ;
	}
      	cerr << " reduced product size : " << (a_spr->size()) << endl;;
	delete a_spr_temp;
	delete a_s;
      } else {
	a_spr = a_s;
      }
	 
    }// for each seed


    if (gv_homogeneous_flag) {
      automaton * na = a_spr->product(*a_homogeneous, true, false);
      delete a_spr;
      a_spr = na;
      if (gv_minimize_flag) {
	automaton * na = a_spr->Hopcroft(); 
	delete a_spr;
	a_spr = na;
      }
    }


    /* (3) compute sensitivity */
    automaton * a_pr_sens     = a_spr->product(a_sens, false, true);    
    double sens               = a_pr_sens->PrFinal(gv_alignment_length)/pr_div_homogeneous;
    delete      a_pr_sens;

    //FIXME double sel = 1.0 - MIN(1.0,gv_nbseeds * selp);
    double sel = 1.0 - MIN(1.0,selp);

    delete a_spr;

    double distance = dist(sel,sens);


    ostringstream outs;
    for (int i = 0 ; i < gv_nbseeds ; i++) {
      outs << (gv_seeds[i])->str();      
      if (gv_cycle_flag)
	outs << "(" << ((gv_cycle_phase[i] - ((gv_seeds[i]->span())%gv_cycle) + gv_cycle)%gv_cycle) << "/" << gv_cycle <<")";
      if (i != gv_nbseeds-1)
	outs << ",";
    }

    seedproperties e = seedproperties(sel,sens,distance,outs.str());
    l.push_back(e);
    
  } else {


    /*
     * B) compute for all sets of seeds ...
     */

    automaton * a_s[MAXSEEDS]     = {NULL};
    double hillclimbing_threshold = -1.00;

    for (int i=0 ; i < gv_nbseeds ; i++)
      gv_seeds[i] = new seed();

    for (int i=0 ; i < gv_nbseeds ; i++)
      gv_cycle_phase[i] = 0;
    
    


    while (1) { /* for all seeds */
      double sens_hillclimbing    = 0.0;
      int    seed_to_hillclimbing = 0;
      int    last_seed_to_hillclimbing = 0;


    sens_and_sel:
     
      /* 1) build seed automata */
      for (int i=0 ; i < gv_nbseeds ; i++) {
	if (!a_s[i]){
	  (a_s[i]) = new automaton();
	  SEEDAUTOMATON((a_s[i]),(gv_seeds[i]),gv_cycle_flag/* merge final states */);	
	  if (gv_minimize_flag) {
	    automaton * na = (a_s[i])->Hopcroft(); 
	    delete (a_s[i]); 
	    (a_s[i]) = na ;
	  }
	}
      }

      /* 2) compute partial selectivity */
      double      selp = 0.0;
      for(int i=0 ; i < gv_nbseeds ; i++){
	automaton * a_pr_s_sel      = (a_s[i])->product(a_sel, false, true);
	selp                        = MAX(selp,a_pr_s_sel->PrFinal(gv_seeds[i]->span()));	
	delete a_pr_s_sel;
      }	
      if (gv_cycle_flag) {
	selp /= gv_cycle;
      }


      /* 3) compute sensitivity */

      
      /* 3.1)  when cycle is allowed (slow since we have to enumerate all possible cycles) */
      double sens = 0.0;
      if (gv_cycle_flag) {
	
	if (gv_cycle_fixedphase_flag) {

	  // do it for a fixed phase
	  automaton * a_spr = NULL;
	  for(int i=0 ; i < gv_nbseeds ; i++) {
	    automaton * a_cycle = new automaton();
	    a_cycle->Automaton_Cycle(gv_cycle,gv_cycle_phase[i]);
	    automaton * a_sc = (a_s[i])->product(*a_cycle,true,false);
	    delete a_cycle;
	    
	    if (i != 0) {
	      automaton * a_spr_temp = a_spr;
	      a_spr =  a_spr->product(*a_sc, false, false);
	      if (gv_minimize_flag) {
		automaton * na = a_spr->Hopcroft(); 
		delete a_spr; 
		a_spr = na ;
	      }
	      delete a_spr_temp;
	      delete a_sc;
	    } else {
	      a_spr = a_sc;
	    }
	  }
	   
	  automaton * a_pr_sens     = a_spr->product(a_sens, false, true);
	  sens                      = a_pr_sens->PrFinal(gv_alignment_length);	
	  delete      a_pr_sens;
	  delete      a_spr;
	  
	    
	} else {

	    // do it for all possible phases
	    for(int p=0 ; p < gv_cycle   ; p++) {
	      automaton * a_spr = NULL;
	      for(int i=0 ; i < gv_nbseeds ; i++) {
		automaton * a_cycle = new automaton();
		a_cycle->Automaton_Cycle(gv_cycle,(gv_cycle_phase[i]+p)%gv_cycle);
		automaton * a_sc = (a_s[i])->product(*a_cycle,true,false);
		delete a_cycle;
		
		if (i != 0) {
		  automaton * a_spr_temp = a_spr;
		  a_spr =  a_spr->product(*a_sc, false, false);
		  if (gv_minimize_flag) {
		    automaton * na = a_spr->Hopcroft(); 
		    delete a_spr; 
		    a_spr = na ;
		  }
		  delete a_spr_temp;
		  delete a_sc;
		} else {
		  a_spr = a_sc;
		}
	      } // for each seed 

	      
	      automaton * a_s_res = a_spr;
	      // homogeneous
	      if (gv_homogeneous_flag) {
		a_s_res = a_spr->product(*a_homogeneous, true, false);
		if (gv_minimize_flag) {
		  automaton * na = a_s_res->Hopcroft(); 
		  delete a_s_res;
		  a_s_res = na;
		}
	      }
	      
	      automaton * a_pr_sens     = a_s_res->product(a_sens, false, true);
	      sens                     += a_pr_sens->PrFinal(gv_alignment_length);	
	      delete      a_pr_sens;
	      delete      a_spr;
	      if (gv_homogeneous_flag) {
		delete a_s_res;
	      }
	    }// for each cycle
	  
	    sens /= gv_cycle;
	    
	} // fixedphase or not

      } else {

	/* 3.2)  no cycle is allowed (faster) */
	automaton * a_spr = NULL;
	for(int i=0 ; i < gv_nbseeds ; i++) {
	  if (i != 0) {
	    automaton * a_spr_temp = a_spr;
	    a_spr =  a_spr->product(*(a_s[i]), false, false);
	    if (gv_minimize_flag) {
	      automaton * na = a_spr->Hopcroft(); 
	      delete a_spr; 
	      a_spr = na ;
	    }
	    if (i > 1)
	      delete a_spr_temp;
	  } else {
	    a_spr = (a_s[0]);
	  }
	} // for each seed 
	
	automaton * a_s_res = a_spr;
	// homogeneous
	if (gv_homogeneous_flag) {
	  a_s_res = a_spr->product(*a_homogeneous, true, false);
	  if (gv_minimize_flag) {
	    automaton * na = a_s_res->Hopcroft(); 
	    delete a_s_res;
	    a_s_res = na;
	  }
	}

	automaton * a_pr_sens     = a_s_res->product(a_sens, false, true);
	sens                      = a_pr_sens->PrFinal(gv_alignment_length);	
	delete      a_pr_sens;
	if (gv_nbseeds > 1)
	  delete a_spr;	
	if (gv_homogeneous_flag) {	  
	  delete a_s_res;
	}
      }
      
      sens /= pr_div_homogeneous;

           
      /* compute selectivity */
      double sel = 1.0 - MIN(1.0,gv_nbseeds*selp);      
      nbruns++;
      
             
      /* (4) build the memorizing structure */
      double distance = dist(sel,sens);      
      ostringstream outs;
      for (int i = 0 ; i < gv_nbseeds ; i++) {
	outs << (gv_seeds[i])->str();     
	if (gv_cycle_flag)	  
	  outs << "(" << ((gv_cycle_phase[i] - ((gv_seeds[i]->span())%gv_cycle) + gv_cycle)%gv_cycle) << "/" << gv_cycle << ")";
	if (i != gv_nbseeds-1)
	  outs << ",";
      }
      seedproperties e = seedproperties(sel,sens,distance,outs.str());



      /* (5) insertion inside pareto set */
      double sensitivity_threshold = insert_Pareto(l,e);
      

      /* (6) Prune list to keep bets ones 
       *     Sleep if requested
       */      
      if (!(nbruns % 1000)) {
	double area = selectPareto(l);
	bool sleeping = false;
      gettime:
	time_t    t    = time(NULL);
	struct tm * tm = localtime(&t);
	if(gv_sleeping_flag && 
	   (tm->tm_hour >= 8 && tm->tm_hour < 20) &&
	   (tm->tm_wday > 0  && tm->tm_wday < 6)
	   ) {
	  if(!sleeping) {
	    /* backup before sleeping */

	    /* security input */
	    if(gv_nb_input_filenames > 0) {
	      for(int i=0;i<gv_nb_input_filenames;i++)
		inputPareto(l,gv_input_filenames[i]);
	    }
	    
	    /* final output */
	    cerr << "# seeds\t sel\t sens\t dist"<<  endl;
	    if (!gv_output_filename)
	      list_and_areaPareto(l);
	    else
	      outputPareto(l,gv_output_filename);
	    cerr << "sleeping ..." << endl;
	    
	    
	    sleeping = true;
	  }
	  SLEEP(600);
	  goto gettime;
	}
	
	cerr << "#runs:"<< nbruns << ",\tset size:" << l.size() << ",\tarea:" << area ;
	cerr << "\t" << (tm->tm_mday) << "/" << (tm->tm_mon) << "/" << (tm->tm_year + 1900) << "\t" << (tm->tm_hour)<<":" << (tm->tm_min) << ":" << (tm->tm_sec);
	cerr << ",\tlast motif checked:" << outs.str() ;
	if  (gv_hillclimbing_flag) {
	  cerr << ",\thillclimbing_threshold:"<< hillclimbing_threshold ;
	}
	cerr << endl;
      }


            
      /* (7) hill climbing method */
      if (gv_hillclimbing_flag) {
	double threshold  = sensitivity_threshold -  hillclimbing_threshold;

#ifdef DEBUG_HILLCLIMBING
	cerr <<  "hillclimbing_threshold (randow walk): " << hillclimbing_threshold ;
#endif

	/* (7.1) set hill_climbing_flag */ 
	bool   hillclimbing_flag   = false;
	if( threshold  >  0.0) {
#ifdef DEBUG_HILLCLIMBING
	  cerr <<  "\t + " << endl;
#endif
	  hillclimbing_flag = true;
	  hillclimbing_threshold = MIN(-1e-6,hillclimbing_threshold + (threshold/sqrt(nbruns))); /* more stringent */
	} else {
#ifdef DEBUG_HILLCLIMBING
	  cerr <<  "\t - " << endl;
#endif
	  hillclimbing_threshold = MAX(-1.00,hillclimbing_threshold + (threshold*gv_hillclimbing_alpha/sqrt(nbruns))); /* little less stringent */
	}
	
	
	/* (7.2) launch hill_climbing procedure */
	if (hillclimbing_flag) {	
	
	  /* hillclimb phenomenom detected */
	  if (sens > sens_hillclimbing) {
#ifdef DEBUG_HILLCLIMBING
	    cerr << "- improved : ";
	    {for (int i = 0 ; i < gv_nbseeds ; i++) {
	      cerr << *(gv_seeds[i]) << ",";
	    }}
	    cerr << "  " << sens << endl; 
#endif

	    sens_hillclimbing      = sens;
	    last_seed_to_hillclimbing = seed_to_hillclimbing;
	    seed_to_hillclimbing ++;  // for the next improving, take the next seed
	    seed_to_hillclimbing %= gv_nbseeds;
	    
	    {for (int i = 0 ; i < gv_nbseeds ; i++) {
	      (gv_seeds[i])->resetswap();
	    }
	    }
	  }
	  

	  /* next seed in hillclimbing order */
	  while ((gv_seeds[seed_to_hillclimbing])->nextswap() == 0 ) {
	    
	    /* delete automaton associated with the modified seed */
	    if (a_s[seed_to_hillclimbing]) {
	      delete (a_s[seed_to_hillclimbing]);
	      a_s[seed_to_hillclimbing] = NULL;
	    }	    
	    seed_to_hillclimbing ++;
	    seed_to_hillclimbing %= gv_nbseeds;
	    if (seed_to_hillclimbing == last_seed_to_hillclimbing) {
	      seed_to_hillclimbing = 0;
	      last_seed_to_hillclimbing = 0;	    
	      goto eoswap;
	    }
	  }
	  

	  /* delete automaton associated with the modified seed */
	  if (a_s[seed_to_hillclimbing]) {
	    delete (a_s[seed_to_hillclimbing]);
	    a_s[seed_to_hillclimbing] = NULL;
	  }
	  
	  goto sens_and_sel;
	eoswap:
	  /* end of hill climbing */
	  sens  = sens_hillclimbing;
	  {
	    for (int i = 0 ; i < gv_nbseeds ; i++) {
	      (gv_seeds[seed_to_hillclimbing])->setswap();
	    }
	  }
	  
#ifdef DEBUG_HILLCLIMBING
	  cerr << "# Final (" << nbruns  << "):" ;
	  {
	    for (int i = 0 ; i < gv_nbseeds ; i++) {
	      cerr << *(gv_seeds[i]) << ",";
	    }
	  }
	  cerr << "\tsel:" << sel << "\tsens:" << sens << endl << endl; 
#endif

	}
      } /* if (gv_hill_climbing_flag)*/
     

  



      
      /* (8) : enumerate the next seed(s) */
      if (gv_nbruns > 0){
	
	double first_seed_weight = 0;
	
	/* (8.1) Random enumeration */
	for (int i = 0 ; i < gv_nbseeds ; i++ ) {
	  
	next:
	  (gv_seeds[i])->random();

	  /* delete automaton associated with the modified seed */
	  if (a_s[i]) {
	    delete (a_s[i]);
	    a_s[i] = NULL;
	  }

	  
	  /* jive selection activated of not  ? */
	  if (gv_jive > 0){ 
	    if (i > 0) {
	      double weight = (gv_seeds[i])->weight();
	      if ( ( weight / first_seed_weight < gv_jive ) || ( first_seed_weight / weight < gv_jive ) ) 
		goto next;
	    } else {
	      first_seed_weight = (gv_seeds[i])->weight();
	      if (first_seed_weight <= 0)
		goto next;
	    }
	  }
	}

	/* phase randomization */	
	if (gv_cycle_flag) {
	  for (int i = 0 ; i < gv_nbseeds ; i++ ) {	
	    gv_cycle_phase[i] = rand()%gv_cycle;
	  }
	}
	
	if (nbruns >= gv_nbruns)
	  goto label_EndOfLoop;
	
      } else { // gv_nbruns > 0
	
	/* (8.2) Complete enumeration */
	int j = 0;
	
	while ( (gv_seeds[j])->next() == 0 ) {
	  
	  /* delete automaton associated with the modified seed */
	  if (a_s[j]) {
	    delete (a_s[j]);
	    a_s[j] = NULL;
	  }
	  
	  j++;
	  if (j == gv_nbseeds) {

	    /* phase enumeration >> */
	    if (gv_cycle_flag) {

	      int k = 0;
	      while (gv_cycle_phase[k] == gv_cycle-1) {
		k++;
		if (k == gv_nbseeds)
		  goto label_EndOfLoop;
	      }
	      
	      gv_cycle_phase[k]++;
	      for(int i = 0 ; i < k ; i++)
		gv_cycle_phase[i] = 0;
	      break;
	    } else {
	      goto label_EndOfLoop;
	    }
	    /* << phase enumeration */
	    
	  }
	} /* while */

	/* delete automaton associated with the modified seed (a_s[j]->next == 1) */
	if (a_s[j]) {
	  delete (a_s[j]);
	  a_s[j] = NULL;
	}

	for(int i = 0 ; i < j ; i++) { // reset the span of previous j-th seeds to min
	  delete (gv_seeds[i]);	  
	  gv_seeds[i] = new seed();
	}

      }/* Random/Complete enum */
    }/* while (1) */
  label_EndOfLoop:  
    ;
  } /* enumerate all seeds /or/ check one seed */




  /*
   * III) Display pareto set and area
   */

  /* update pareto set according to other processes (helpful on parallel computation) */
  if(gv_nb_input_filenames > 0) {
    for(int i=0;i<gv_nb_input_filenames;i++)
      inputPareto(l,gv_input_filenames[i]);
  }
  
  /* final output */
  cerr << "# seeds\t sel\t sens\t dist"<<  endl;
  if (!gv_output_filename)
    list_and_areaPareto(l);
  else
    outputPareto(l,gv_output_filename);

  return 0;
}



